EXTERN          _main
GLOBAL          _go16, _go32, _base, _mode_info, _vbe_fn5
GLOBAL          _move_window, _memset, _memcpy, _beep, _nobeep, _img
GLOBAL          _v86_code_return, _mw_esp, _v86_move_window_end

[SECTION .text]
[BITS 32]

_go16:

        incbin  "a16.bin"

[BITS 16]

        jmp     dword 8:_go32                   ; jump to 32-bit code

[BITS 32]

_go32:
        xor     eax, eax
        mov     fs, eax
        mov     gs, eax
        mov     ax, 10h
        mov     ds, eax
        mov     es, eax
        mov     ss, eax

        mov     [_base], ebx
        mov     [_mode_info], edx

        or      edi, edi
        jz      .1                              ; skip conversion if no PMode 
                                                ; interface but v86 is required
        sub     edi, ebx                        ; convert entry point to near
.1:
        mov     [_vbe_fn5], edi

        in      al, 0a1h
        mov     ah, al
        in      al, 21h
        mov     [_irq_mask], eax                ; save IRQ masks

        xor     al, al
        dec     al
        out     21h, al
        out     0A1h, al                        ; disable all IRQs

        push    ecx
        call    _main
        pop     ecx

        mov     eax, [_irq_mask]
        out     21h, al
        mov     al, ah
        out     0A1h, al                        ; restore IRQ masks

        jmp     dword 18h:_go16back             ; jump to 16-bit code

_go16back:

[BITS 16]

        mov     ax, 20h
        mov     ds, ax
        mov     es, ax
        mov     ss, ax
        mov     fs, ax
        mov     gs, ax

        mov     eax, cr0
        and     al, 0feh
        mov     cr0, eax

        push    cx
        push    word 100h+_16cont-_go16
        retf

[BITS 32]
_16cont:
[BITS 16]

        mov     ds, cx
        mov     es, cx
        mov     ss, cx

        sti

        push    byte 40h
        pop     es
        mov     cx, 180                 ; approx. 10 seconds delay
mtwait:
        mov     ax, [es:6ch]
mttick:
        cmp     ax, [es:6ch]
        je      mttick
        loop    mtwait                  ; wait for a while...

        mov     ax, 3
        int     10h			; restore color 80x25 text mode

        mov     ax, 4c00h
        int     21h

[BITS 32]

%define PARAM_1 ebp+8+4*0
%define PARAM_2 ebp+8+4*1
%define PARAM_3 ebp+8+4*2
%define PARAM_4 ebp+8+4*3

_move_window:
        mov     eax, [_vbe_fn5]
        or      eax, eax
        jz      _v86_move_window

        push    ebp
        mov     ebp, esp
        push    es
        pushad

        mov     ax, 4f05h
        xor     ebx, ebx
        mov     edx, [PARAM_1]
        call    [_vbe_fn5]
        mov     ax, 4f05h
        xor     ebx, ebx
        inc     ebx
        mov     edx, [PARAM_1]
        call    [_vbe_fn5]

        popad
        pop     es
        pop     ebp
        ret

_v86_move_window:
        push    ebp
        mov     ebp, esp
        pushad
        mov     [_mw_esp], esp

        mov     ax, [PARAM_1]
        shl     eax, 16
        mov     ax, [PARAM_1]
        push    eax                     ; put granule on v68 stack twice

        mov     edx, esp                ; this would be v86 esp...

        xor     eax, eax
        push    eax                     ; gs
        push    eax                     ; fs
        push    eax                     ; ds
        push    eax                     ; es

        mov     ecx, [_base]
        shr     ecx, 4
        push    ecx                     ; ss

        push    edx                     ; esp

        push    dword 23002h            ; eflags (vm=1,iopl=3,if=0)
        push    ecx                     ; cs
        push    dword _v86_code         ; eip

        iretd
_v86_move_window_end:

        popad
        pop     ebp
        ret

_v86_code:
[BITS 16]
        mov     ax, 4f05h
        xor     bx, bx
        pop     dx
                                        ; btw, ds = 0 here :)
        pushf                           ; this is a lot faster than "int 10h"
        call    far [40h]               ; because "int" causes the exception

        mov     ax, 4f05h
        xor     bx, bx
        inc     bx
        pop     dx
                                        ; btw, ds = 0 here :)
        pushf                           ; this is a lot faster than "int 10h"
        call    far [40h]               ; because "int" causes the exception
[BITS 32]
_v86_code_return:
        int     0

_memset:
        push    ebp
        mov     ebp, esp
        push    edi
        mov     edi, [PARAM_1]
        mov     al, [PARAM_2]
        mov     ecx, [PARAM_3]
        rep     stosb
        mov     eax, [PARAM_1]
        pop     edi
        pop     ebp
        ret

_memcpy:
        push    ebp
        mov     ebp, esp
        push    esi
        push    edi
        mov     edi, [PARAM_1]
        mov     esi, [PARAM_2]
        mov     ecx, [PARAM_3]
        mov     eax, ecx
        shr     ecx, 2
        rep     movsd
        and     eax, byte 3
        mov     ecx, eax
        rep     movsb
        mov     eax, [PARAM_1]
        pop     edi
        pop     esi
        pop     ebp
        ret

_beep:
        xor     edx, edx
        mov     ebx, 5000

        in      al, 61h
        or      al, 3
        out     61h, al
        mov     al, 0b6h
        out     43h, al

        mov     eax, 1190000
        div     ebx

        out     42h, al
        mov     al, ah
        out     42h, al
        ret

_nobeep:
        in      al, 61h
        and     al, 0fch
        out     61h, al
        ret

[section .data]
_img    incbin  "image.img"

[section .bss]
_base           resd 1
_mode_info      resd 1
_vbe_fn5        resd 1
_mw_esp         resd 1
_irq_mask       resd 1
