GLOBAL  _isr_01_wrapper, _isr_0D_wrapper, _isr_0E_wrapper

EXTERN  _single_step_handler, _gpf_handler, _page_fault_handler
EXTERN  _base, _mw_esp, _v86_move_window_end, _v86_code_return

[SECTION .text]
[BITS 32]

_isr_01_wrapper:
        push    ds
        push    es
        push    fs
        push    gs                      ; saving segment registers and
        pushad                          ; other regs because it's an ISR
        mov     bx, 10h
        mov     ds, ebx
        mov     es, ebx                 ; load ds and es with valid selector
        call    _single_step_handler    ; call actual ISR code
        popad                           ; restoring the regs
        pop     gs
        pop     fs
        pop     es
        pop     ds
        and     dword [esp+8], 0FFFFFEFFh; EFLAGS.TF = 0
        iretd

_isr_0D_wrapper:
        test    dword [esp+12], 20000h
        jnz     .1                      ; if v86 code
.0:
        xchg    ecx, [esp+8]            ;-) exchange ecx and old CS
        xchg    edx, [esp+4]            ;-) exchange edx and old EIP
        xchg    eax, [esp]              ;-) push eax / mov eax, [error code]
        push    ds
        push    es
        push    fs
        push    gs                      ; saving segment registers and
        pushad                          ; other regs because it's an ISR
        mov     bx, 10h
        mov     ds, ebx
        mov     es, ebx                 ; load ds and es with valid selector
        push    ecx                     ; pass old CS
        push    edx                     ; pass old EIP
        push    eax                     ; pass error code as a parameter
        call    _gpf_handler            ; call actual ISR code
        pop     eax
        pop     eax
        pop     eax
        popad                           ; restoring the regs
        pop     gs
        pop     fs
        pop     es
        pop     ds
        xchg    ecx, [esp+8]            ;-) exchange ecx and old CS
        xchg    edx, [esp+4]            ;-) exchange edx and old EIP
        pop     eax
        iretd
.1:
        xchg    eax, [esp+20]           ;-) exchange eax and old SS
        xchg    ebx, [esp+16]           ;-) exchange ebx and old ESP
        xchg    edi, [esp+12]           ;-) exchange edi and EFLAGS
        xchg    ecx, [esp+8]            ;-) exchange ecx and old CS
        xchg    edx, [esp+4]            ;-) exchange edx and old EIP
        push    ebp
        push    esi

        mov     si, 10h
        mov     ds, esi
        mov     es, esi                 ; load ds and es with valid selector
        mov     si, 28h
        mov     gs, esi                 ; gs = 4gb seg selector

        movzx   eax, ax
        movzx   ebx, bx
        movzx   ecx, cx
        movzx   edx, dx

        mov     ebp, [_base]
        shr     ebp, 4                  ; ebp = .COM program segment

        cmp     ecx, ebp
        jne     .2                      ; cs <> seg
        cmp     edx, _v86_code_return
        jne     .2                      ; ip <> _v86_code_return

        mov     eax, ds
        mov     ss, eax
        mov     esp, [_mw_esp]          ; use regular stack instead of stack#0
        jmp     _v86_move_window_end    ; finally return from v86

.2:        
        mov     esi, ecx
        shl     esi, 4
        add     esi, edx                ; esi points to the instruction
        cmp     byte [gs:esi], 0cdh
        je      .3                      ; if "int n"

        pop     esi
        pop     ebp
        xchg    eax, [esp+20]           ;-) exchange eax and old SS
        xchg    ebx, [esp+16]           ;-) exchange ebx and old ESP
        xchg    edi, [esp+12]           ;-) exchange edi and EFLAGS
        xchg    ecx, [esp+8]            ;-) exchange ecx and old CS
        xchg    edx, [esp+4]            ;-) exchange edx and old EIP
        jmp     .0

.3:

        add     edx, byte 2             ; adjust IP to next instruction

        mov     ebp, eax
        shl     ebp, 4
        add     ebp, ebx                ; ebp = ss*16+sp of v86

        sub     ebp, byte 2
        mov     [gs:ebp], di            ; push FLAGS
        sub     ebp, byte 2
        mov     [gs:ebp], cx            ; push CS
        sub     ebp, byte 2
        mov     [gs:ebp], dx            ; push IP

        sub     ebx, byte 6             ; adjust SP

        inc     esi
        movzx   esi, byte [gs:esi]      ; esi = int number
        shl     esi, 2
        mov     dx, [gs:esi]            ; IP = offset of ISR
        mov     cx, [gs:esi+2]          ; CS = segment of ISR

        and     di, 0fcffh              ; reset IF and TF in EFLAGS

        pop     esi
        pop     ebp
        xchg    eax, [esp+20]           ;-) exchange eax and old SS
        xchg    ebx, [esp+16]           ;-) exchange ebx and old ESP
        xchg    edi, [esp+12]           ;-) exchange edi and EFLAGS
        xchg    ecx, [esp+8]            ;-) exchange ecx and old CS
        xchg    edx, [esp+4]            ;-) exchange edx and old EIP
        add     esp, byte 4             ; get a rid of the error code
        iretd                           ; continue in real mode ISR

_isr_0E_wrapper:
        xchg    eax, [esp]              ;-) push eax / mov eax, [error code]
        push    ds
        push    es
        push    fs
        push    gs                      ; saving segment registers and
        pushad                          ; other regs because it's an ISR
        mov     bx, 10h
        mov     ds, ebx
        mov     es, ebx                 ; load ds and es with valid selector
        push    eax                     ; pass error code as a parameter
        call    _page_fault_handler     ; call actual ISR code
        or      eax, eax
        pop     eax
        popad                           ; restoring the regs
        pop     gs
        pop     fs
        pop     es
        pop     ds
        pop     eax
        jz      .1                      ; don't set single-step flag
        or      dword [esp+8], 100h     ; EFLAGS.TF = 1
.1:
        iretd
