#ifndef _pm_
#define _pm_

#include "types.h"

/***************************/
/* PMode-related constants */
/***************************/

/*****************************/
/* Segment-related constants */
/*****************************/

#define ACS_PRESENT     0x80
#define ACS_CODE        0x18
#define ACS_DATA        0x10
#define ACS_EXP_DOWN    0x04
#define ACS_CONFORM     0x04
#define ACS_READABLE    0x02
#define ACS_WRITEABLE   0x02

#define ACS_CODE_SEG    (ACS_CODE | ACS_PRESENT)
#define ACS_DATA_SEG    (ACS_DATA | ACS_PRESENT | ACS_WRITEABLE)
#define ACS_STACK_SEG   ACS_DATA_SEG

#define ACS_IDT         ACS_DATA_SEG
#define ACS_INT         0x0E
#define ACS_TRAP        0x0F
#define ACS_INT_GATE    (ACS_INT | ACS_PRESENT)
#define ACS_TRAP_GATE   (ACS_TRAP | ACS_PRESENT)
#define ACS_TSS         (0x09 | ACS_PRESENT)

#define ACS_DPL_0       0x00
#define ACS_DPL_1       0x20
#define ACS_DPL_2       0x40
#define ACS_DPL_3       0x60

#define ATTR_32BIT      0x40
#define ATTR_BIG        0x40
#define ATTR_GRANUL     0x80

#define RPL_0           0x00
#define RPL_1           0x01
#define RPL_2           0x02
#define RPL_3           0x03

/**************************/
/* Page-related constants */
/**************************/

#define PAGE_SIZE       0x1000
#define PAGE_MASK       0xFFFFF000
#define PAGE_TABLE_ENTRIES 0x400

#define P_PRESENT       0x01
#define P_READWRITE     0x02
#define P_USERSUPER     0x04

/****************************/
/* PMode-related structures */
/****************************/

#pragma pack (push, 1)

typedef struct {
  word limit;
  word base0_15;
  byte base16_23;
  byte attrs0_7;
  byte attrs8_15;
  byte base24_31;
} SEG_DESCRIPTOR;

typedef struct {
  word offs0_15;
  word sel;
  byte paramcnt;
  byte attrs;
  word offs16_31;
} INT_DESCRIPTOR;

typedef struct {
  word limit;
  dword base;
} GDT_REG;

typedef struct {
  word limit;
  dword base;
} IDT_REG;

typedef struct {
  dword link,
        esp0,
        ss0,
        esp1,
        ss1,
        esp2,
        ss2,
        cr3,
        eip,
        eflags,
        eax,
        ecx,
        edx,
        ebx,
        esp,
        ebp,
        esi,
        edi,
        es,
        cs,
        ss,
        ds,
        fs,
        gs,
        ldtr;
  word  trace,
        io_map_addr;
  byte  io_map[8192];
} TSS;

#pragma pack (pop)

/*****************************/
/* PMode-related subroutines */
/*****************************/

void LGDT (GDT_REG *gdtr);
void SGDT (GDT_REG *gdtr);

void LIDT (IDT_REG *idtr);
void SIDT (IDT_REG *idtr);

void LTR (dword sel);
dword STR();

void LCR3 (dword value);
dword SCR3();

dword SCR2();

void CLTS();

void INVLPG (void *mem);

void RESET();

#endif
